#######################################################################
#
# spma component
#
# generates the SPMA configuration file, runs SPMA if required.
#
# German Cancio <German.Cancio@cern.ch>
# Version 1.2.8 : 25/11/04 15:17
#
# Copyright (c) 2003 German Cancio, CERN and EU DataGrid.
# For license conditions see http://www.eu-datagrid.org/license.html
#
#######################################################################

package NCM::Component::spma;
#
# a few standard statements, mandatory for all components
#
use strict;
use NCM::Component;
use vars qw(@ISA $EC);
@ISA = qw(NCM::Component);
$EC=LC::Exception::Context->new->will_store_all;

use NCM::Check;
use LC::Process;


# SPMA config file
my $spma_cfg='/var/lib/spma-target.cf';

# unescape by defaut
my $dounescape=1;
my $trailprefix=1;


#
# small helper function for unescaping chars
#
sub unescape ($) {
  my $str=shift;

  if ($dounescape==0) {
    if ($trailprefix ==1 && $str =~ m%^_%) {
      return substr($str,1);
    } else {
      return $str;
    }
  }

  $str =~ s!(_[0-9a-f]{2})!sprintf("%c",hex($1))!eg;
  return $str;
}


#
# get_repositories($config):ref(%hash)
#
# get a reference to a hash of hashes with the used repositories
#
# structure:
# $rep{repname}{protocol} = url
# eg.
# $rep{'CERN_CC'}{'http'} = 'http://swrep.cern.ch/swrep'
#
#
sub get_repositories($) {
  my ($self,$config)=@_;

  #
  # get the repositories
  #
  my %repository=();

  my $path='/software/repositories';

  my $rep_list=$config->getElement($path);
  unless (defined $rep_list) {
    $EC->ignore_error();
    $self->error('cannot access config path: '.$path);
    return undef;
  }

  my ($rep_el,$rep_el_name,$rep_name,$rep_name_str,$url);
  my ($rep_prot,$rep_prot_entry,$rep_prot_entry_name,$rep_prot_url);
  my ($rep_prot_name,$rep_prot_name_str);

  #
  # the NVA API navigation functions would profit from
  # shortcut methods for relative navigation,
  # something like
  #
  # $newelement=$oldelement->down('relative/path')
  #
  # and probably also
  #
  # $newelvalue=$oldelement->downvalue('relative/path');
  #
  # instead of having to instantiate a new element from
  # $config
  #

  while($rep_list->hasNextElement()) {
    $rep_el=$rep_list->getNextElement();
    $rep_el_name=$rep_el->getName();
    $rep_name=$config->getElement($path.'/'.$rep_el_name.'/name');
    $rep_name_str=$rep_name->getValue();

    $rep_prot=$config->getElement($path.'/'.$rep_el_name.'/protocols');
    while ($rep_prot->hasNextElement()) {
      $rep_prot_entry=$rep_prot->getNextElement();
      $rep_prot_entry_name=$rep_prot_entry->getName();
      $rep_prot_name=$config->getElement
        ($path.'/'.$rep_el_name.'/protocols/'.
         $rep_prot_entry_name.'/name');
      $rep_prot_url=$config->getElement
        ($path.'/'.$rep_el_name.'/protocols/'.
         $rep_prot_entry_name.'/url');
      $repository{$rep_name_str}{$rep_prot_name->getValue()}=
        $rep_prot_url->getValue();
    }
  }
  return \%repository;
}

#
# get_packages($config,$rep):ref(@array)
#
# returns a reference to an array with the list of
# target packages.
# each element of the array is a reference to a hash:
# 'rep'     => repository name (eg. 'CERN_CC')
# 'name'    => package name    (eg. 'emacs')
# 'version' => version-release (eg. '20.4-3')
# 'arch'    => architecture    (eg. i386)
#
# 'flags'   => flags ref(%hash) with flagtype=>value ('reboot'=>'true')
#
sub get_packages ($) {
  my ($self,$config,$rep)=@_;

  my @pkg_list=();

  my $error=0;
  my $path='/software/packages';

  my ($pkg_el,$pkg_name,$pkg_verlist,$pkgver,$pkgver_name);
  my ($pkgver_arch,$pkgver_rep,$pkg_flaglist,$pkg_flag,$use_rep);

  my $pkg_list=$config->getElement($path);
  while ($pkg_list->hasNextElement()) {
    $pkg_el=$pkg_list->getNextElement();
    $pkg_name=$pkg_el->getName();
    #
    # go for the versions
    #
    $pkg_verlist=$config->getElement($path.'/'.$pkg_name);

    while($pkg_verlist->hasNextElement()) {
      $pkgver=$pkg_verlist->getNextElement();
      $pkgver_name=$pkgver->getName();
      $pkgver_arch=$config->getElement($path.'/'.$pkg_name.
                                       '/'.$pkgver_name.'/arch');
     if (!defined $pkgver_arch || $pkgver_arch->getValue() eq '') {
	$EC->ignore_error();
	$self->error('cannot get architecture for '.
		     &unescape($pkg_name).'-'.&unescape($pkgver_name));
	$error++;
	next;
      }
      $pkgver_rep=$config->getElement($path.'/'.$pkg_name.
                                      '/'.$pkgver_name.'/repository');
      if (!defined $pkgver_rep || $pkgver_rep->getValue() eq '') {
	$EC->ignore_error();
	#
	# if more than one repository available, give up, otherwise
	# assume unique repository is the one to use
	#
	my @reps=keys %{$rep};
	if (scalar @reps >1) {
	  $self->error('cannot guess repository for '.
		       &unescape($pkg_name).'-'.&unescape($pkgver_name));
	  $error++;
	  next;
        } else {
	  $use_rep=$reps[0];
	}
      } else {
	$use_rep=$pkgver_rep->getValue();
      }
      #
      # get the per package flags, if any
      #
      my %flags=();
      unless ($pkg_flaglist=$config->getElement($path.'/'.$pkg_name.
                                               '/'.$pkgver_name.'/flags')) {
	$EC->ignore_error();
      } else {
	while ($pkg_flaglist->hasNextElement()) {
	  $pkg_flag=$pkg_flaglist->getNextElement();
	  $flags{$pkg_flag->getName()}=$pkg_flag->getValue();
	}
      }

      push(@pkg_list,
           {'rep'     => $use_rep,
            'name'    => &unescape($pkg_name),
            'version' => &unescape($pkgver_name),
            'arch'    => $pkgver_arch->getValue(),
            'flags'   => \%flags,
           }
          );
    }
  }
  return $error ? undef : \@pkg_list;
}


#
# run the SPMA
#
sub run_spma {
  my ($self)=@_;

  my $spma_exec="/usr/bin/spma --quiet";
  $self->report('running the SPMA: '.$spma_exec);

  my ($stdout,$stderr);
  my $execute_status = LC::Process::execute([$spma_exec],
                                              timeout => 8000,
                                              stdout => \$stdout,
                                              stderr => \$stderr
                                              );

  unless (defined $execute_status) {
    $self->error("could not execute $spma_exec");
    return;
  }
  if ($?) {
    $self->error("$spma_exec failed: $? (please check spma.log)");
  } else {
    $self->OK ("$spma_exec run succesfully (please check spma.log)");
  }
  if ($stdout) {
    $self->warn("$spma_exec output produced: (please check spma.log)");
    $self->report($stdout);
  }
  if ($stderr) {
    $self->warn("$spma_exec STDERR output produced: (please check spma.log)");
    $self->report($stderr);
  }
}

#
# function for updating the spma.conf configuration file
#

sub updateconffile {
  my ($self,$config)=@_;

  my @keys=qw(userpkgs userprio packager usespmlist rpmexclusive debug verbose
              cachedir localcache proxy proxytype proxyhost proxyport headnode);

  my $conffile='/etc/spma.conf';
  unless (-e $conffile && -w $conffile) {
    $self->warn('does not exist or cannot write to: '.$conffile);
    return;
  }

  my $i;
  my %key;
  foreach $i (@keys) {
    my $cfgel=$config->getElement('/software/components/spma/'.$i);
    unless (defined $cfgel) {
      $EC->ignore_error();
      next;
    } else {
      $key{$i}=$cfgel->getValue();
    }
  }
  #
  # if using a headnode, then set proxyhost to it
  #
  if (exists $key{'headnode'} && $key{'headnode'} eq 'true') {
    unless ($config->elementExists('/hardware/serial_console/headnode')) {
      $self->warn('head node activated, but none found in /hardware/serial_console/headnode');
      $key{'proxy'}="no";
      $key{'proxyhost'}='undefined';
  } else {
      $key{'proxyhost'}=$config->getValue('/hardware/serial_console/headnode');
    }
    delete $key{'headnode'}; # as not part of SPMA config file
  }
  foreach $i (keys %key) {
    NCM::Check::lines($conffile,
		    backup => ".old",
		      linere => '^#?\s*'.$i.'(\s+|=).*',
		      goodre => '^\s*'.$i.'\s*=\s*'.$key{$i},
		      good   => "$i = ".$key{$i},
		      add => 'last'
		     );
  }
}






##########################################################################
sub Configure {
##########################################################################
  my ($self,$config)=@_;


  $self->updateconffile($config);

  # do we need to escape chars?
  if ($config->elementExists('/software/components/spma/unescape') &&
      $config->getValue('/software/components/spma/unescape') eq 'false') {
    $dounescape=0;
  }

  # do we need to remove a trailing underscore? (only if no unescape())
  if ($config->elementExists('/software/components/spma/trailprefix') &&
      $config->getValue('/software/components/spma/trailprefix') eq 'false') {
    $trailprefix=0;
  }


  my $rep=$self->get_repositories($config);
  unless (defined $rep) {
    $self->error ('could not get repository information');
    return;
  }

  my $pkgs=$self->get_packages($config,$rep);

  unless (defined $pkgs) {
    $self->error('error generating package list - preserving old one');
    return;
  }

  #
  # build the config file
  #

  local(*CFG);

  unless ($NoAction) {
    unless(open(CFG,">$spma_cfg.$$")) {
      $self->error ("cannot open $spma_cfg.$$");
      return;
    }
  }

  print CFG "#\n#\n# generated by NCM SPMA component at ".scalar(localtime)."\n#\n#\n"
    unless ($NoAction);
  my ($i,$ver,$rel);
  foreach $i (@$pkgs) {
    # SPMA wants separate version and release
    ($ver,$rel)=split('-',$i->{'version'});
    unless (exists($rep->{$i->{'rep'}}{'http'})) {
      $self->error ("no HTTP protocol found for repository ".$i->{'rep'});
      unlink ("$spma_cfg.$$");
      return;
      # this is arbitrary. In a next version,
      # there should be the possibility of specifying
      # which protocols are acceptable (ordered list)
    }

    #
    # flags
    #
    my (@flags,$f);
    foreach $f (keys %{$i->{'flags'}}) {
      if ($i->{'flags'}->{$f} eq 'true') {
        push(@flags,uc($f));
      }
    }

    print CFG $rep->{$i->{'rep'}}{'http'} .' '.$i->{'name'}.
              ' '.$ver.' '.$rel.' '.$i->{'arch'}.' '.
                join(' ',@flags)."\n"
		  unless ($NoAction);
  }

  unless ($NoAction) {
    unless(close(CFG)) {
      $self->error("cannot close $spma_cfg.$$");
      return;
    }
    unless (rename ("$spma_cfg.$$",$spma_cfg)) {
      $self->error("cannot rename temporary $spma_cfg.$$ to $spma_cfg");
      return;
    }
  }
  $self->OK("updated SPMA target configuration file in $spma_cfg");

  if ($config->elementExists('/software/components/spma/run') &&
      $config->getValue('/software/components/spma/run') eq 'yes') {
    $self->run_spma();
  }

  return;
}

1; # required for Perl modules
